/*  This file is part of distributed convex BP - http://www.alexander-schwing.de/
 *
 *  distributed convex BP is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  distributed convex BP is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with distributed convex BP.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Copyright (C) 2010-2011  Alexander Schwing  [aschwing _at_ inf _dot_ ethz _dot_ ch]
 */

#include "Client.h"

#include <mpi.h>

#include <iostream>
#include <fstream>
#include <algorithm>
#include <cmath>
#include <cstdlib>
#include <string.h>

#include <omp.h>

#define _DEF_ABS(x) (((x)<0) ? -(x) : (x))

#define DEL_ARRAY(name)										if((name)!=NULL) {		\
																delete [] (name);	\
																(name) = NULL;		\
															}

template <class T>
Client<T>::Client() {
	s_nodePotentialIX = NULL;
	s_factorPotentialIX = NULL;
	CumPosNodesToFactors = NULL;
	CumPosFactorsToNodes = NULL;
	Ni = NULL;
	Nalpha = NULL;
	s_NodePotSize = NULL;
	NodePosInFactor = NULL;
	c_i = NULL;
	hat_c_alpha = NULL;
	s_NodePotPosition = NULL;
	s_NodePotentials = NULL;
	s_FactorPotPosition = NULL;
	s_FactorVariableSizePosition = NULL;
	s_FactorPotentials = NULL;
	s_FactorVariableSize = NULL;
	s_DividedFactors = NULL;
	CumPosMsg = NULL;	
	log_n = NULL;
	log_ns = NULL;

	CumPosFactorBelief = NULL;
	CumPosNodeBelief = NULL;

	//in case hyperthreading is enabled, you might want to consider restricting OpenMP to the number of physically available cores such that you don't harm performance
	//omp_set_num_threads(8);
}

template <class T>
Client<T>::~Client() {
	DEL_ARRAY(s_nodePotentialIX);
	DEL_ARRAY(s_factorPotentialIX);
	DEL_ARRAY(CumPosNodesToFactors);
	DEL_ARRAY(CumPosFactorsToNodes);
	DEL_ARRAY(Ni);
	DEL_ARRAY(Nalpha);
	DEL_ARRAY(s_NodePotSize);
	DEL_ARRAY(NodePosInFactor);
	DEL_ARRAY(c_i);
	DEL_ARRAY(hat_c_alpha);
	DEL_ARRAY(s_NodePotPosition);
	DEL_ARRAY(s_NodePotentials);
	DEL_ARRAY(s_FactorPotPosition);
	DEL_ARRAY(s_FactorVariableSizePosition);
	DEL_ARRAY(s_FactorPotentials);
	DEL_ARRAY(s_FactorVariableSize);
	DEL_ARRAY(s_DividedFactors);
	DEL_ARRAY(CumPosMsg);	
	DEL_ARRAY(log_n);
	DEL_ARRAY(log_ns);

	DEL_ARRAY(CumPosFactorBelief);
	DEL_ARRAY(CumPosNodeBelief);
}

template <class T>
T Client<T>::computeM(size_t i, size_t alpha, size_t x_i) {
	size_t factorPotIX = s_factorPotentialIX[alpha];
	size_t factorPotPos = s_FactorPotPosition[factorPotIX];
	T* theta_alpha = &s_FactorPotentials[factorPotPos];

	size_t i_localIX = size_t(-1);
	for(size_t iIX=CumPosFactorsToNodes[alpha];iIX<CumPosFactorsToNodes[alpha+1];++iIX) {
		if(Nalpha[iIX]==i) {
			i_localIX = NodePosInFactor[iIX];
			break;
		}
	}

	size_t numVarsInFactor = s_FactorVariableSizePosition[factorPotIX+1]-s_FactorVariableSizePosition[factorPotIX];

	size_t* vars = new size_t[numVarsInFactor];
	vars[i_localIX] = x_i;

	size_t* FactorVariableSize = &s_FactorVariableSize[s_FactorVariableSizePosition[factorPotIX]];

	size_t threshold = 1;
	size_t* prodStates = new size_t[numVarsInFactor+1];
	prodStates[0] = 1;
	for(size_t iIX=s_FactorVariableSizePosition[factorPotIX];iIX<s_FactorVariableSizePosition[factorPotIX+1];++iIX) {
		size_t var = iIX-s_FactorVariableSizePosition[factorPotIX];
		prodStates[var+1] = s_FactorVariableSize[iIX]*prodStates[var];
		threshold *= (i_localIX==var) ? 1 : s_FactorVariableSize[iIX];
	}

	T* mem = new T[threshold];
	T maxVal = T(0.0);

	for(size_t x_alphaNOi=0;x_alphaNOi<threshold;++x_alphaNOi) {
		size_t potentIX = 0;
		for(size_t m=0;m<numVarsInFactor;++m) {
			if(m<i_localIX) {
				vars[m] = (x_alphaNOi/prodStates[m])%FactorVariableSize[m];
			} else if(m>i_localIX) {
				vars[m] = ((x_alphaNOi*s_NodePotSize[i])/prodStates[m])%FactorVariableSize[m];
			}
			potentIX += vars[m]*prodStates[m];
		}

		mem[x_alphaNOi] = theta_alpha[potentIX];

		if(s_DividedFactors[alpha]!=size_t(-1)) {	//this is a clique that is divided
			mem[x_alphaNOi] += log_ns[s_DividedFactors[alpha]+potentIX];
		}

		for(size_t jIX=CumPosFactorsToNodes[alpha];jIX<CumPosFactorsToNodes[alpha+1];++jIX) {
			size_t j=Nalpha[jIX];
			if(j==i) {
				continue;
			}

			size_t local_alphaIX=CumPosNodesToFactors[j];
			for(;local_alphaIX!=CumPosNodesToFactors[j+1];++local_alphaIX) {
				if(Ni[local_alphaIX]==alpha) {
					break;
				}
			}
			mem[x_alphaNOi] += (vars[NodePosInFactor[jIX]]==0) ? T(0.0) : log_n[CumPosMsg[j] + (local_alphaIX-CumPosNodesToFactors[j])*(s_NodePotSize[j]-1) + (vars[NodePosInFactor[jIX]]-1)];
		}

		maxVal = (mem[x_alphaNOi]>maxVal || x_alphaNOi==0) ? mem[x_alphaNOi] : maxVal;
	}

	T ec = epsilon*hat_c_alpha[alpha];
	if(ec!=T(0.0)) {							
		T sumVal = exp( (mem[0] - maxVal)/ec );
		for(size_t x_alphaNOi=1;x_alphaNOi<threshold;++x_alphaNOi) {
			sumVal += exp( (mem[x_alphaNOi] - maxVal)/ec );
		}
		maxVal += ec*log(sumVal);
	}

	delete [] vars;
	delete [] mem;
	delete [] prodStates;

	return maxVal;
}

template <class T>
void Client<T>::computeN(size_t i) {
	size_t nodePotIX = s_nodePotentialIX[i];
	T* theta_i;
	if(nodePotIX==size_t(-1)) {
		theta_i = NULL;
	} else {
		size_t nodePotPos = s_NodePotPosition[nodePotIX];
		theta_i = &s_NodePotentials[nodePotPos];
	}

	T* m_alphaTOi = new T[CumPosNodesToFactors[i+1]-CumPosNodesToFactors[i]];
	T* sum_log_n_x_0 = new T[CumPosNodesToFactors[i+1]-CumPosNodesToFactors[i]];

	for(size_t x_i=0;x_i<s_NodePotSize[i];++x_i) {
		T tmp = (theta_i==NULL) ? 0 : theta_i[x_i];
		T hatci = c_i[i];

		for(size_t alphaIX=CumPosNodesToFactors[i];alphaIX<CumPosNodesToFactors[i+1];++alphaIX) {
			size_t alpha = Ni[alphaIX];

			hatci += hat_c_alpha[alpha];

			size_t tmpvar = alphaIX-CumPosNodesToFactors[i];

			m_alphaTOi[tmpvar] = computeM(i,alpha,x_i);
			tmp += m_alphaTOi[tmpvar];
		}

		tmp /= hatci;

		for(size_t alphaIX=CumPosNodesToFactors[i];alphaIX<CumPosNodesToFactors[i+1];++alphaIX) {
			size_t alpha = Ni[alphaIX];

			size_t tmpvar = alphaIX-CumPosNodesToFactors[i];

			if(x_i==0) {
				sum_log_n_x_0[tmpvar] = tmp*hat_c_alpha[alpha] - m_alphaTOi[tmpvar];
			} else {
				log_n[CumPosMsg[i]+tmpvar*(s_NodePotSize[i]-1) + (x_i-1)] = tmp*hat_c_alpha[alpha] - m_alphaTOi[tmpvar] - sum_log_n_x_0[tmpvar];
			}
		}
	}

	delete [] m_alphaTOi;
	delete [] sum_log_n_x_0;
}


template <class T>
int Client<T>::DoComputation() {
	for(size_t iteration=0;iteration<num_minor_iterations;++iteration) {
		for(size_t k=0;k<Colors.size();++k) {
#pragma omp parallel
{
#pragma omp for nowait
			for(OPENMP_ITER_TYPE m=0;m<OPENMP_ITER_TYPE(Colors[k].size());++m) {
				computeN(Colors[k][m]);
			}
}//end parallel section
		}
	}

	return 0;
}

template <class T>
void Client<T>::computeLocalBelief(T* localBelief) {
#pragma omp parallel
{
#pragma omp for nowait
	for(OPENMP_ITER_TYPE i=0;i<OPENMP_ITER_TYPE(num_Nodes);++i) {
		size_t nodePotIX = s_nodePotentialIX[i];
		T* theta_i;
		if(nodePotIX==size_t(-1)) {
			theta_i = NULL;
		} else {
			size_t nodePotPos = s_NodePotPosition[nodePotIX];
			theta_i = &s_NodePotentials[nodePotPos];
		}

		T* localBelief_i = localBelief;
		for(size_t k=0;k<size_t(i);++k) {
			localBelief_i += s_NodePotSize[k];
		}

		T maxVal = T(0.0);
		for(size_t x_i=0;x_i<s_NodePotSize[i];++x_i) {
			localBelief_i[x_i] = (theta_i==NULL) ? 0 : theta_i[x_i];

			for(size_t alphaIX=CumPosNodesToFactors[i];alphaIX<CumPosNodesToFactors[i+1];++alphaIX) {

				localBelief_i[x_i] -= (x_i==0) ? T(0.0) : log_n[CumPosMsg[i]+(alphaIX-CumPosNodesToFactors[i])*(s_NodePotSize[i]-1) + (x_i-1)];

			}

			maxVal = (localBelief_i[x_i]>maxVal || x_i==0) ? localBelief_i[x_i] : maxVal;
		}

		T ec = epsilon*c_i[i];
		if(ec==T(0.0)) {
			T sum = 0.0f;
			for(size_t x_i=0;x_i<s_NodePotSize[i];++x_i) {
				localBelief_i[x_i] = (localBelief_i[x_i] - maxVal)>-1e-5 ? T(1.0) : T(0.0);
				sum += localBelief_i[x_i];
			}
			for(size_t x_i=0;x_i<s_NodePotSize[i];++x_i) {
				localBelief_i[x_i] /= sum;
			}
		} else {
			T sum = T(0.0);
			for(size_t x_i=0;x_i<s_NodePotSize[i];++x_i) {
				sum += exp( (localBelief_i[x_i] - maxVal)/ec );
			}
			for(size_t x_i=0;x_i<s_NodePotSize[i];++x_i) {
				localBelief_i[x_i] = exp( (localBelief_i[x_i] - (maxVal + ec*log(sum)) )/ec );
			}
		}
	}
}//end parallel section
}

template <class T>
void Client<T>::computeFactorBelief(T* factorBelief) {
#pragma omp parallel
{
#pragma omp for nowait
	for(OPENMP_ITER_TYPE alpha=0;alpha<OPENMP_ITER_TYPE(num_Factors);++alpha) {
		size_t factorPotIX = s_factorPotentialIX[alpha];
		size_t factorPotPos = s_FactorPotPosition[factorPotIX];
		T* theta_alpha = &s_FactorPotentials[factorPotPos];

		size_t numVarsInFactor = s_FactorVariableSizePosition[factorPotIX+1]-s_FactorVariableSizePosition[factorPotIX];

		size_t* vars = new size_t[numVarsInFactor];

		size_t* FactorVariableSize = &s_FactorVariableSize[s_FactorVariableSizePosition[factorPotIX]];

		size_t threshold = 1;
		size_t* prodStates = new size_t[numVarsInFactor+1];
		prodStates[0] = 1;
		for(size_t iIX=s_FactorVariableSizePosition[factorPotIX];iIX<s_FactorVariableSizePosition[factorPotIX+1];++iIX) {
			prodStates[iIX-s_FactorVariableSizePosition[factorPotIX]+1] = s_FactorVariableSize[iIX]*prodStates[iIX-s_FactorVariableSizePosition[factorPotIX]];
			threshold *= s_FactorVariableSize[iIX];
		}

		T maxVal = T(0.0);

		T* buf = factorBelief + CumPosFactorBelief[alpha];

		for(size_t x_alpha=0;x_alpha<threshold;++x_alpha) {
			size_t potentIX = 0;
			for(size_t m=0;m<numVarsInFactor;++m) {
				vars[m] = (x_alpha/prodStates[m])%FactorVariableSize[m];
				potentIX += vars[m]*prodStates[m];
			}

			buf[x_alpha] = theta_alpha[potentIX];

			if(s_DividedFactors[alpha]!=size_t(-1)) {	//this is a clique that is divided
				buf[x_alpha] += log_ns[s_DividedFactors[alpha]+potentIX];
			}

			for(size_t iIX=CumPosFactorsToNodes[alpha];iIX<CumPosFactorsToNodes[alpha+1];++iIX) {
				size_t i=Nalpha[iIX];

				size_t local_alphaIX=CumPosNodesToFactors[i];
				for(;local_alphaIX!=CumPosNodesToFactors[i+1];++local_alphaIX) {
					if(Ni[local_alphaIX]==size_t(alpha)) {
						break;
					}
				}
				buf[x_alpha] += (vars[NodePosInFactor[iIX]]==0) ? T(0.0) : log_n[CumPosMsg[i] + (local_alphaIX-CumPosNodesToFactors[i])*(s_NodePotSize[i]-1) + (vars[NodePosInFactor[iIX]]-1)];
			}

			maxVal = (buf[x_alpha]>maxVal || x_alpha==0) ? buf[x_alpha] : maxVal;
		}

		T ec = epsilon*hat_c_alpha[alpha];
		if(ec==T(0.0)) {
			T sum = T(0.0);
			for(size_t x_alpha=0;x_alpha<threshold;++x_alpha) {
				buf[x_alpha] = (buf[x_alpha] - maxVal)>-1e-5 ? T(1.0) : T(0.0);
				sum += buf[x_alpha];
			}
			for(size_t x_alpha=0;x_alpha<threshold;++x_alpha) {
				buf[x_alpha] /= sum;
			}
		} else {
			T sum = T(0.0);
			for(size_t x_alpha=0;x_alpha<threshold;++x_alpha) {
				sum += exp( (buf[x_alpha]-maxVal)/ec );
			}
			for(size_t x_alpha=0;x_alpha<threshold;++x_alpha) {
				buf[x_alpha]  = exp( (buf[x_alpha] - (maxVal + ec*log(sum)))/ec );
			}
		}
		
		delete [] vars;
		delete [] prodStates;

	}
}//end parallel section
}

template <class T>
void Client<T>::checkAgreement(bool* agreement, T* factorBelief, T* localBelief) {
#pragma omp parallel
{
#pragma omp for nowait
	for(OPENMP_ITER_TYPE alpha=0;alpha<OPENMP_ITER_TYPE(num_Factors);++alpha) {
		size_t factorPotIX = s_factorPotentialIX[alpha];

		T* factorBelief_alpha = factorBelief + CumPosFactorBelief[alpha];

		size_t numVarsInFactor = s_FactorVariableSizePosition[factorPotIX+1]-s_FactorVariableSizePosition[factorPotIX];

		size_t* vars = new size_t[numVarsInFactor];

		size_t* FactorVariableSize = &s_FactorVariableSize[s_FactorVariableSizePosition[factorPotIX]];

		size_t threshold = 1;
		size_t* prodStates = new size_t[numVarsInFactor+1];
		prodStates[0] = 1;
		for(size_t iIX=s_FactorVariableSizePosition[factorPotIX];iIX<s_FactorVariableSizePosition[factorPotIX+1];++iIX) {
			prodStates[iIX-s_FactorVariableSizePosition[factorPotIX]+1] = s_FactorVariableSize[iIX]*prodStates[iIX-s_FactorVariableSizePosition[factorPotIX]];
			threshold *= s_FactorVariableSize[iIX];
		}

		agreement[alpha] = true;
		
		for(size_t iIX=CumPosFactorsToNodes[alpha];iIX<CumPosFactorsToNodes[alpha+1];++iIX) {
			size_t i = Nalpha[iIX];

			T* NodeBelief = new T[s_NodePotSize[i]];
			memset((void*)NodeBelief, 0, sizeof(T)*s_NodePotSize[i]);

			for(size_t x_alpha=0;x_alpha<threshold;++x_alpha) {
				size_t potentIX = 0;
				for(size_t m=0;m<numVarsInFactor;++m) {
					vars[m] = (x_alpha/prodStates[m])%FactorVariableSize[m];
					potentIX += vars[m]*prodStates[m];
				}

				NodeBelief[vars[NodePosInFactor[iIX]]] += factorBelief_alpha[potentIX];
			}

			T* localBelief_i = localBelief + CumPosNodeBelief[i];

			for(size_t x_i=0;x_i<s_NodePotSize[i];++x_i) {
				if(_DEF_ABS(localBelief_i[x_i]-NodeBelief[x_i])>agreementThreshold) {
					agreement[alpha] = false;
					break;
				}
			}

			delete [] NodeBelief;

			if(agreement[alpha] == false) {
				break;
			}
		}

		delete [] vars;
		delete [] prodStates;
	}
}//end parallel section
}

template <class T>
T Client<T>::computePrimal(T* localBelief) {
	double primalCost = 0.0;

	computeLocalBelief(localBelief);

#pragma omp parallel
{
	double primalCost_threadLocal = 0.0;
#pragma omp for nowait
	for(OPENMP_ITER_TYPE i=0;i<OPENMP_ITER_TYPE(num_Nodes);++i) {
		size_t nodePotIX = s_nodePotentialIX[i];
		T* theta_i;
		if(nodePotIX==size_t(-1)) {
			theta_i = NULL;
		} else {
			size_t nodePotPos = s_NodePotPosition[nodePotIX];
			theta_i = &s_NodePotentials[nodePotPos];
		}

		T* localBelief_i = localBelief;
		for(size_t k=0;k<size_t(i);++k) {
			localBelief_i += s_NodePotSize[k];
		}

		double tmp = 0.0;
		if(theta_i!=NULL) {
			for(size_t x_i=0;x_i<s_NodePotSize[i];++x_i) {
				tmp += theta_i[x_i]*localBelief_i[x_i];			
			}
			primalCost_threadLocal += tmp;
		}

		T ec = epsilon*c_i[i];
		if(ec!=T(0.0)) {
			tmp = 0.0;
			for(size_t x_i=0;x_i<s_NodePotSize[i];++x_i) {
				T tmpval = localBelief_i[x_i];
				tmp -= ec*tmpval*log(tmpval + 1e-50);
			}
			primalCost_threadLocal += tmp;
		}
	}
#pragma omp critical
{
	primalCost += primalCost_threadLocal;
}//end critical section
}//end parallel section

	size_t factorBeliefSize = 0;
	for(size_t alpha=0;alpha<num_Factors;++alpha) {
		size_t fPIX = s_factorPotentialIX[alpha];
		size_t fS = 1;
		for(size_t m=s_FactorVariableSizePosition[fPIX];m<s_FactorVariableSizePosition[fPIX+1];++m) {
			fS *= s_FactorVariableSize[m];
		}
		factorBeliefSize += fS;
	}

	T* factorBelief = new T[factorBeliefSize];
	computeFactorBelief(factorBelief);

	bool* agreement = new bool[num_Factors];
	checkAgreement(agreement, factorBelief, localBelief);

	num_agrees = 0;
	for(size_t k=0;k<num_Factors;++k) {
		if(agreement[k])
			++num_agrees;
	}

#pragma omp parallel
{
	double primalCost_threadLocal = 0.0;
#pragma omp for nowait
	for(OPENMP_ITER_TYPE alpha=0;alpha<OPENMP_ITER_TYPE(num_Factors);++alpha) {
		size_t factorPotIX = s_factorPotentialIX[alpha];
		size_t factorPotPos = s_FactorPotPosition[factorPotIX];
		T* theta_alpha = &s_FactorPotentials[factorPotPos];

		size_t numVarsInFactor = s_FactorVariableSizePosition[factorPotIX+1]-s_FactorVariableSizePosition[factorPotIX];

		size_t* vars = new size_t[numVarsInFactor];

		size_t* FactorVariableSize = &s_FactorVariableSize[s_FactorVariableSizePosition[factorPotIX]];

		T* factorBelief_alpha = factorBelief + CumPosFactorBelief[alpha];

		size_t threshold = 1;
		size_t* prodStates = new size_t[numVarsInFactor+1];
		prodStates[0] = 1;
		for(size_t iIX=s_FactorVariableSizePosition[factorPotIX];iIX<s_FactorVariableSizePosition[factorPotIX+1];++iIX) {
			prodStates[iIX-s_FactorVariableSizePosition[factorPotIX]+1] = s_FactorVariableSize[iIX]*prodStates[iIX-s_FactorVariableSizePosition[factorPotIX]];
			threshold *= s_FactorVariableSize[iIX];
		}

		T ec = epsilon*hat_c_alpha[alpha];
		double tmp = 0.0;
		for(size_t x_alpha=0;x_alpha<threshold;++x_alpha) {
			size_t potentIX = 0;
			for(size_t m=0;m<numVarsInFactor;++m) {
				vars[m] = (x_alpha/prodStates[m])%FactorVariableSize[m];
				potentIX += vars[m]*prodStates[m];
			}

			double buf = 1.0;

			if(!agreement[alpha]) {
				for(size_t iIX=CumPosFactorsToNodes[alpha];iIX<CumPosFactorsToNodes[alpha+1];++iIX) {
					size_t i = Nalpha[iIX];

					T* localBelief_i = localBelief + CumPosNodeBelief[i];

					buf *= localBelief_i[vars[NodePosInFactor[iIX]]];
				}
			} else {
				buf *= factorBelief_alpha[potentIX];
			}
			if(ec!=T(0.0)) {
				tmp -= ec*buf*log(buf+1e-50);
			}
			tmp += buf*theta_alpha[potentIX];
		}

		primalCost_threadLocal += tmp;

		delete [] vars;
		delete [] prodStates;
	}
#pragma omp critical
{
	primalCost += primalCost_threadLocal;
}//end critical section
}//end parallel section

	delete [] factorBelief;
	delete [] agreement;

	return T(primalCost);
}

template <class T>
T Client<T>::computeDual() {
	double dualCost = 0.0;

#pragma omp parallel
{
	double dualCost_threadLocal = 0.0;	
#pragma omp for nowait
	for(OPENMP_ITER_TYPE i=0;i<OPENMP_ITER_TYPE(num_Nodes);++i) {
		size_t nodePotIX = s_nodePotentialIX[i];
		T* theta_i;
		if(nodePotIX==size_t(-1)) {
			theta_i = NULL;
		} else {
			size_t nodePotPos = s_NodePotPosition[nodePotIX];
			theta_i = &s_NodePotentials[nodePotPos];
		}
		T* buf = new T[s_NodePotSize[i]];

		double maxVal = 0.0;
		for(size_t x_i=0;x_i<s_NodePotSize[i];++x_i) {
			buf[x_i] = (theta_i!=NULL) ? theta_i[x_i] : 0;

			for(size_t alphaIX=CumPosNodesToFactors[i];alphaIX<CumPosNodesToFactors[i+1];++alphaIX) {

				buf[x_i] -= (x_i==0) ? T(0.0) : log_n[CumPosMsg[i]+(alphaIX-CumPosNodesToFactors[i])*(s_NodePotSize[i]-1) + (x_i-1)];

			}

			maxVal = (double(buf[x_i])>maxVal || x_i==0) ? double(buf[x_i]) : maxVal;
		}

		T ec = epsilon*c_i[i];
		if(ec!=T(0.0)) {
			double sum = 0.0;
			for(size_t x_i=0;x_i<s_NodePotSize[i];++x_i) {
				sum += exp( (buf[x_i] - maxVal)/ec );
			}
			maxVal += ec*log(sum);
		}

		dualCost_threadLocal += maxVal;
		delete [] buf;
	}
#pragma omp critical
{
	dualCost += dualCost_threadLocal;
}//end critical section
}//end parallel section


#pragma omp parallel
{
	double dualCost_threadLocal = 0.0;
#pragma omp for nowait
	for(OPENMP_ITER_TYPE alpha=0;alpha<OPENMP_ITER_TYPE(num_Factors);++alpha) {
		size_t factorPotIX = s_factorPotentialIX[alpha];
		size_t factorPotPos = s_FactorPotPosition[factorPotIX];
		T* theta_alpha = &s_FactorPotentials[factorPotPos];

		size_t numVarsInFactor = s_FactorVariableSizePosition[factorPotIX+1]-s_FactorVariableSizePosition[factorPotIX];

		size_t* vars = new size_t[numVarsInFactor];

		size_t* FactorVariableSize = &s_FactorVariableSize[s_FactorVariableSizePosition[factorPotIX]];

		size_t threshold = 1;
		size_t* prodStates = new size_t[numVarsInFactor+1];
		prodStates[0] = 1;
		for(size_t iIX=s_FactorVariableSizePosition[factorPotIX];iIX<s_FactorVariableSizePosition[factorPotIX+1];++iIX) {
			prodStates[iIX-s_FactorVariableSizePosition[factorPotIX]+1] = s_FactorVariableSize[iIX]*prodStates[iIX-s_FactorVariableSizePosition[factorPotIX]];
			threshold *= s_FactorVariableSize[iIX];
		}

		T* mem = new T[threshold];
		double maxVal = 0.0;

		for(size_t x_alpha=0;x_alpha<threshold;++x_alpha) {
			size_t potentIX = 0;
			for(size_t m=0;m<numVarsInFactor;++m) {
				vars[m] = (x_alpha/prodStates[m])%FactorVariableSize[m];
				potentIX += vars[m]*prodStates[m];
			}
			
			mem[x_alpha] = theta_alpha[potentIX];

			if(s_DividedFactors[alpha]!=size_t(-1)) {	//this is a clique that is divided
				mem[x_alpha] += log_ns[s_DividedFactors[alpha]+potentIX];
			}

			for(size_t iIX=CumPosFactorsToNodes[alpha];iIX<CumPosFactorsToNodes[alpha+1];++iIX) {
				size_t i=Nalpha[iIX];

				size_t local_alphaIX=CumPosNodesToFactors[i];
				for(;local_alphaIX!=CumPosNodesToFactors[i+1];++local_alphaIX) {
					if(Ni[local_alphaIX]==size_t(alpha)) {
						break;
					}
				}
				mem[x_alpha] += (vars[NodePosInFactor[iIX]]==0) ? T(0.0) : log_n[CumPosMsg[i] + (local_alphaIX-CumPosNodesToFactors[i])*(s_NodePotSize[i]-1) + (vars[NodePosInFactor[iIX]]-1)];
			}

			maxVal = (mem[x_alpha]>maxVal || x_alpha==0) ? double(mem[x_alpha]) : maxVal;
		}

		T ec = epsilon*hat_c_alpha[alpha];
		if(ec!=T(0.0)) {
			double sum = 0.0;
			for(size_t x_alpha=0;x_alpha<threshold;++x_alpha) {
				sum += exp( (mem[x_alpha]-maxVal)/ec );
			}
			maxVal += ec*log(sum);
		}
		
		delete [] mem;
		delete [] prodStates;
		delete [] vars;		

		dualCost_threadLocal += maxVal;
	}
#pragma omp critical
{
	dualCost += dualCost_threadLocal;
}//end critical section
}//end parallel section

	return T(dualCost);
}

template <class T>
int Client<T>::GreedyColoring() {
	size_t* ColorOfNode = new size_t[num_Nodes];
	memset(ColorOfNode, 0, sizeof(size_t)*num_Nodes);
	std::vector<size_t> CurrentlyUsedColors;
	for(size_t i=0;i<num_Nodes;++i) {
		std::vector<size_t> NeighborColors;

		for(size_t alphaIX=CumPosNodesToFactors[i];alphaIX<CumPosNodesToFactors[i+1];++alphaIX) {
			size_t alpha = Ni[alphaIX];
			
			for(size_t iIX=CumPosFactorsToNodes[alpha];iIX<CumPosFactorsToNodes[alpha+1];++iIX) {
				size_t j=Nalpha[iIX];

				if(j!=i && ColorOfNode[j]!=0) {
					NeighborColors.push_back(ColorOfNode[j]);
				}
			}
		}

		bool ColorAssigned = false;
		for(size_t k=0;k<CurrentlyUsedColors.size();++k) {
			std::vector<size_t>::iterator pos = std::find(NeighborColors.begin(), NeighborColors.end(), CurrentlyUsedColors[k]);
			if(pos==NeighborColors.end()) {
				ColorOfNode[i] = CurrentlyUsedColors[k];
				ColorAssigned = true;
				break;
			}
		}
		if(!ColorAssigned) {
			size_t newcol = CurrentlyUsedColors.size()+1;
			CurrentlyUsedColors.push_back(newcol);
			ColorOfNode[i] = newcol;
		}
	}

	Colors.assign(CurrentlyUsedColors.size(), std::vector<size_t>());
	for(size_t i=0;i<num_Nodes;++i) {
		Colors[ColorOfNode[i]-1].push_back(i);
	}

	delete [] ColorOfNode;

	return 0;
}

template <class T>
int Client<T>::ReceiveProblem(int myid) {
	ClientID = myid;

	MPI::Status stat;

	MPI::COMM_WORLD.Probe(0, 0, stat);
	int count = stat.Get_count(MPI::CHAR);

	char* buf = new char[count];

	MPI::COMM_WORLD.Recv(buf, count, MPI::CHAR, 0, 0, stat);

	size_t* tmp = (size_t*)buf;
	num_Nodes = *tmp++;
	num_Factors = *tmp++;
	num_minor_iterations = *tmp++;

	size_t s_NodePotPositionSize = *tmp++;
	size_t s_FactorPotPositionSize = *tmp++;
	size_t s_FactorVariableSizePositionSize = *tmp++;

	if(sizeof(T)>sizeof(size_t)) {
		float epsilonTmp = *((float*)tmp);
		epsilon = T(epsilonTmp);
	} else {
		epsilon = *((T*)tmp);
	}

	++tmp;

	if(sizeof(T)>sizeof(size_t)) {
		float f_athreshold = *((float*)tmp);
		agreementThreshold = T(f_athreshold);
	} else {
		agreementThreshold = *((T*)tmp);
	}

	delete [] buf;

	//std::cout << num_nodes << "\t" << num_factors << "\t" << num_minor_iterations << "\t" << num_node_pot << "\t" << num_factor_pot << "\t" << num_return_solutions << "\t" << cBPparams.epsilon << std::endl;

	s_nodePotentialIX = new size_t[num_Nodes];
	MPI::COMM_WORLD.Recv((char*)s_nodePotentialIX, int(num_Nodes*sizeof(size_t)), MPI::CHAR, 0, 1, stat);

	s_factorPotentialIX = new size_t[num_Factors];
	MPI::COMM_WORLD.Recv((char*)s_factorPotentialIX, int(num_Factors*sizeof(size_t)), MPI::CHAR, 0, 2, stat);

	CumPosNodesToFactors = new size_t[num_Nodes+1];
	MPI::COMM_WORLD.Recv((char*)CumPosNodesToFactors, int(sizeof(size_t)*(num_Nodes+1)), MPI::CHAR, 0, 3, stat);

	CumPosFactorsToNodes = new size_t[num_Factors+1];
	MPI::COMM_WORLD.Recv((char*)CumPosFactorsToNodes, int(sizeof(size_t)*(num_Factors+1)), MPI::CHAR, 0, 4, stat);

	Ni = new size_t[CumPosNodesToFactors[num_Nodes]];
	MPI::COMM_WORLD.Recv((char*)Ni, int(sizeof(size_t)*CumPosNodesToFactors[num_Nodes]), MPI::CHAR, 0, 5, stat);

	Nalpha = new size_t[CumPosFactorsToNodes[num_Factors]];
	MPI::COMM_WORLD.Recv((char*)Nalpha, int(sizeof(size_t)*CumPosFactorsToNodes[num_Factors]), MPI::CHAR, 0, 6, stat);

	s_NodePotSize = new size_t[num_Nodes];
	MPI::COMM_WORLD.Recv((char*)s_NodePotSize, int(sizeof(size_t)*num_Nodes), MPI::CHAR, 0, 7, stat);

	NodePosInFactor = new size_t[CumPosFactorsToNodes[num_Factors]];
	MPI::COMM_WORLD.Recv((char*)NodePosInFactor, int(sizeof(size_t)*CumPosFactorsToNodes[num_Factors]), MPI::CHAR, 0, 8, stat);

	c_i = new T[num_Nodes];
	MPI::COMM_WORLD.Recv((char*)c_i, int(sizeof(T)*num_Nodes), MPI::CHAR, 0, 9, stat);

	hat_c_alpha = new T[num_Factors];
	MPI::COMM_WORLD.Recv((char*)hat_c_alpha, int(sizeof(T)*num_Factors), MPI::CHAR, 0, 10, stat);

	s_NodePotPosition = new size_t[s_NodePotPositionSize];
	MPI::COMM_WORLD.Recv((char*)s_NodePotPosition, int(sizeof(size_t)*s_NodePotPositionSize), MPI::CHAR, 0, 11, stat);

	s_NodePotentials = new T[s_NodePotPosition[s_NodePotPositionSize-1]];
	MPI::COMM_WORLD.Recv((char*)s_NodePotentials, int(sizeof(T)*s_NodePotPosition[s_NodePotPositionSize-1]), MPI::CHAR, 0, 12, stat);

	s_FactorPotPosition = new size_t[s_FactorPotPositionSize];
	MPI::COMM_WORLD.Recv((char*)s_FactorPotPosition, int(sizeof(size_t)*s_FactorPotPositionSize), MPI::CHAR, 0, 13, stat);

	s_FactorVariableSizePosition = new size_t[s_FactorVariableSizePositionSize];
	MPI::COMM_WORLD.Recv((char*)s_FactorVariableSizePosition, int(sizeof(size_t)*s_FactorVariableSizePositionSize), MPI::CHAR, 0, 14, stat);

	s_FactorPotentials = new T[s_FactorPotPosition[s_FactorPotPositionSize-1]];
	MPI::COMM_WORLD.Recv((char*)s_FactorPotentials, int(sizeof(T)*s_FactorPotPosition[s_FactorPotPositionSize-1]), MPI::CHAR, 0, 15, stat);

	s_FactorVariableSize = new size_t[s_FactorVariableSizePosition[s_FactorVariableSizePositionSize-1]];
	MPI::COMM_WORLD.Recv((char*)s_FactorVariableSize, int(sizeof(size_t)*s_FactorVariableSizePosition[s_FactorVariableSizePositionSize-1]), MPI::CHAR, 0, 16, stat);

	s_DividedFactors = new size_t[num_Factors+1];
	MPI::COMM_WORLD.Recv((char*)s_DividedFactors, int(sizeof(size_t)*(num_Factors+1)), MPI::CHAR, 0, 17, stat);

	//reserve message space
	CumPosMsg = new size_t[num_Nodes+1];
	CumPosMsg[0] = 0;
	for(size_t i=0;i<num_Nodes;++i) {
		CumPosMsg[i+1] = (CumPosNodesToFactors[i+1]-CumPosNodesToFactors[i])*(s_NodePotSize[i]-1) + CumPosMsg[i];
	}
	log_n = new T[CumPosMsg[num_Nodes]];
	memset(log_n,0,CumPosMsg[num_Nodes]*sizeof(T));

	size_t log_ns_Size = s_DividedFactors[num_Factors];
	if(log_ns_Size>0) {
		log_ns = new T[log_ns_Size];
		memset(log_ns,0,log_ns_Size*sizeof(T));
	}

	GreedyColoring();

	CumPosFactorBelief = new size_t[num_Factors+1];
	CumPosFactorBelief[0] = 0;
	for(size_t alpha=0;alpha<num_Factors;++alpha) {
		size_t fPIX = s_factorPotentialIX[alpha];
		size_t fS = 1;
		for(size_t m=s_FactorVariableSizePosition[fPIX];m<s_FactorVariableSizePosition[fPIX+1];++m) {
			fS *= s_FactorVariableSize[m];
		}
		CumPosFactorBelief[alpha+1] = fS + CumPosFactorBelief[alpha];
	}

	CumPosNodeBelief = new size_t[num_Nodes+1];
	CumPosNodeBelief[0] = 0;
	for(size_t i=0;i<num_Nodes;++i) {
		CumPosNodeBelief[i+1] = s_NodePotSize[i] + CumPosNodeBelief[i];
	}

	return 0;
}

template <class T>
int Client<T>::SendMsg() {

	std::vector<size_t> Idx;
	std::vector<T> Msg;

	Msg.push_back(computeDual());
	
	size_t localBeliefSize = 0;
	for(size_t i=0;i<num_Nodes;++i) {
		localBeliefSize += s_NodePotSize[i];
	}
	T* localBelief = new T[localBeliefSize];

	Msg.push_back(computePrimal(localBelief));
	Msg.push_back(T(num_agrees));
	
	for(size_t alpha=0;alpha<num_Factors;++alpha) {
		if(s_DividedFactors[alpha]!=size_t(-1)) {	//this is a clique that is divided
			for(size_t iIX=CumPosFactorsToNodes[alpha];iIX<CumPosFactorsToNodes[alpha+1];++iIX) {
				size_t i=Nalpha[iIX];

				size_t local_alphaIX=CumPosNodesToFactors[i];
				for(;local_alphaIX!=CumPosNodesToFactors[i+1];++local_alphaIX) {
					if(Ni[local_alphaIX]==alpha) {
						break;
					}
				}

				Idx.push_back(alpha);
				Idx.push_back(i);

				Msg.push_back(T(0.0));
				for(size_t x_i=1;x_i<s_NodePotSize[i];++x_i) {
					Msg.push_back(log_n[CumPosMsg[i]+(local_alphaIX-CumPosNodesToFactors[i])*(s_NodePotSize[i]-1) + (x_i-1)]);
				}
			}			
		}
	}

	delete [] localBelief;

	if(Idx.size()==0) {
		Idx.push_back(-1);
	}

	MPI::COMM_WORLD.Send((char*)&Idx[0], int(sizeof(size_t)*Idx.size()), MPI::CHAR, 0, 0);
	MPI::COMM_WORLD.Send((char*)&Msg[0], int(sizeof(T)*Msg.size()), MPI::CHAR, 0, 1);
	
	return 0;
}

template <class T>
int Client<T>::ProcessUpdate() {
	MPI::Status stat;
	MPI::COMM_WORLD.Probe(0, 1, stat);
	int count = stat.Get_count(MPI::CHAR);

	char* UpdateMsg = new char[count];
	MPI::COMM_WORLD.Recv(UpdateMsg, count, MPI::CHAR, 0, 1, stat);

	if(count==1 && UpdateMsg[0]=='\0') {
		
	} else {
		if(s_DividedFactors[num_Factors]>0) {
			delete [] log_ns;
			log_ns = (T*)UpdateMsg;
		} else {
			std::cout << "This is not reasonable..." << std::endl;
			delete [] UpdateMsg;
		}
	}

	return 0;
}

template <class T>
int Client<T>::ReceiveUpdateOrBeliefRequest() {
	char what[7];
	MPI::Status stat;

	MPI::COMM_WORLD.Recv(what, 7, MPI::CHAR, 0, 0, stat);

	if(strcmp(what,"Update")==0) {
		ProcessUpdate();
		return 0;
	} else {
		ComputeAndSendLocalBeliefs();
		return 1;
	}

	return -1;
}

template <class T>
int Client<T>::ComputeAndSendLocalBeliefs() {
	size_t localBeliefSize = 0;
	for(size_t i=0;i<num_Nodes;++i) {
		localBeliefSize += s_NodePotSize[i];
	}
	T* localBelief = new T[localBeliefSize];

	computeLocalBelief(localBelief);

	MPI::COMM_WORLD.Send((char*)localBelief, int(sizeof(T)*localBeliefSize), MPI::CHAR, 0, 0);

	delete [] localBelief;

	return 0;
}

template class Client<double>;
template class Client<float>;