/*  This file is part of distributed convex BP - http://www.alexander-schwing.de/
 *
 *  distributed convex BP is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  distributed convex BP is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with distributed convex BP.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Copyright (C) 2010-2011  Alexander Schwing  [aschwing _at_ inf _dot_ ethz _dot_ ch]
 */

#include "Server.h"

#include <mpi.h>

#include <algorithm>
#include <cmath>
#include <iostream>
#include <fstream>
#include <cstdlib>
#include <string.h>

template <class T>
Server<T>::Server(std::vector<std::vector<size_t> >* NToF, std::vector<std::vector<size_t> >* FToN, std::vector<PotentialStruct<T> >* nodePotentials, std::vector<PotentialStruct<T> >* factorPotentials, std::vector<size_t>* nodePotentialIX, std::vector<size_t>* factorPotentialIX, std::vector<size_t>* variableSize, struct cBP_Settings<double>* params, std::vector<size_t>* NodesOnMachine, std::vector<T>* c_i, std::vector<T>* c_alpha) {
	this->m_numNodes = NToF->size();
	this->nodeTo = NToF;				
	this->factorTo = FToN;				
	this->nodePotentials = nodePotentials;	
	this->factorPotentials = factorPotentials;
	this->nodePotentialIX = nodePotentialIX;					
	this->factorPotentialIX = factorPotentialIX;
	this->variableSize = variableSize;
	this->NodesOnMachine = NodesOnMachine;
	this->in_c_i = c_i;
	this->in_c_alpha = c_alpha;

	this->m_numMachines = (*std::max_element(NodesOnMachine->begin(), NodesOnMachine->end())) + 1;

	AllS_DividedFactors.resize(m_numMachines);
	AllS_OldFactorIndexToNew.resize(m_numMachines);
	AllS_CumPosFactorsToNodes.resize(m_numMachines);
	AllS_Nalpha.resize(m_numMachines);
	AllS_NodePotSize.resize(m_numMachines);
	AllS_NodePosInFactor.resize(m_numMachines);
	AllS_OldNodeIndexToNew.resize(m_numMachines);
	
	this->cBPparams = params;

	//N_p(\alpha)
	Npalpha.resize(factorTo->size());
	for(size_t k=0;k<Npalpha.size();++k) {
		for(size_t m=0;m<(*factorTo)[k].size();++m) {
			Npalpha[k].insert((*NodesOnMachine)[(*factorTo)[k][m]]);
		}
	}
}

template <class T>
Server<T>::~Server() {

}

template <class T>
int Server<T>::SendProblemToClient(size_t k) {
	size_t FactorIndexCount = 0;
	size_t NodeIndexCount = 0;
	std::map<size_t, size_t> OldFactorIndexToNew;
	std::map<size_t, size_t> OldNodeIndexToNew;
	std::vector<size_t> s_NodePotSize;

	//first we generate maps that map the original node and factor indices to new indices on machine k
	for(size_t i=0;i<nodeTo->size();++i) {		//loop over all nodes of the graphical model
		if((*NodesOnMachine)[i]==k) {
			OldNodeIndexToNew.insert(std::make_pair<size_t, size_t>(i, NodeIndexCount++));
			s_NodePotSize.push_back((*variableSize)[i]);

			/*for(size_t m=0;m<(*nodeTo)[i].size();++m) {		//loop over all factors this node is connected to
				if(OldFactorIndexToNew.find((*nodeTo)[i][m])==OldFactorIndexToNew.end()) {		//we don't yet have this factor on this machine
					OldFactorIndexToNew.insert(std::make_pair<size_t, size_t>((*nodeTo)[i][m], FactorIndexCount++));
				}
			}*/
		}
	}

	for(size_t alpha=0;alpha<factorTo->size();++alpha) {
		for(size_t i=0;i<(*factorTo)[alpha].size();++i) {
			if((*NodesOnMachine)[(*factorTo)[alpha][i]]==k) {
				OldFactorIndexToNew.insert(std::make_pair<size_t, size_t>(alpha, FactorIndexCount++));
				break;
			}
		}
	}

	//now we generate a flattened N(i) using CumPosNodesToFactors for indexing
	std::vector<size_t> CumPosNodesToFactors(OldNodeIndexToNew.size()+1,0);
	CumPosNodesToFactors[0] = 0;
	for(std::map<size_t, size_t>::const_iterator citer=OldNodeIndexToNew.begin();citer!=OldNodeIndexToNew.end();++citer) {
		CumPosNodesToFactors[citer->second+1] = (*nodeTo)[citer->first].size();
	}
	for(size_t cnt=1;cnt<CumPosNodesToFactors.size();++cnt) {
		CumPosNodesToFactors[cnt] += CumPosNodesToFactors[cnt-1];
	}

	std::vector<size_t> tmpNodeAdd(OldNodeIndexToNew.size(),0);
	std::vector<size_t> Ni(CumPosNodesToFactors.back());
	for(std::map<size_t, size_t>::const_iterator citer=OldNodeIndexToNew.begin();citer!=OldNodeIndexToNew.end();++citer) {
		for(std::vector<size_t>::const_iterator citer_alpha=(*nodeTo)[citer->first].begin();citer_alpha!=(*nodeTo)[citer->first].end();++citer_alpha) {
			Ni[CumPosNodesToFactors[citer->second]+tmpNodeAdd[citer->second]++] = OldFactorIndexToNew.find(*citer_alpha)->second;
		}
	}

	//now we generate a flattened N(alpha) \cup s using CumPosFactorsToNodes for indexing
	std::vector<size_t> CumPosFactorsToNodes(OldFactorIndexToNew.size()+1,0);
	CumPosFactorsToNodes[0] = 0;
	for(std::map<size_t, size_t>::const_iterator citer=OldFactorIndexToNew.begin();citer!=OldFactorIndexToNew.end();++citer) {
		//count how many nodes are on machine k
		size_t absNalphaCs = 0;
		for(std::vector<size_t>::const_iterator citer_i=(*factorTo)[citer->first].begin();citer_i!=(*factorTo)[citer->first].end();++citer_i) {
			if((*NodesOnMachine)[*citer_i]==k) {
				++absNalphaCs;
			}
		}		
		CumPosFactorsToNodes[citer->second+1] = absNalphaCs;
	}
	for(size_t cnt=1;cnt<CumPosFactorsToNodes.size();++cnt) {
		CumPosFactorsToNodes[cnt] += CumPosFactorsToNodes[cnt-1];
	}

	std::vector<size_t> tmpFactorAdd(OldFactorIndexToNew.size(),0);
	std::vector<size_t> Nalpha(CumPosFactorsToNodes.back());
	std::vector<size_t> NodePosInFactor(CumPosFactorsToNodes.back());
	for(std::map<size_t, size_t>::const_iterator citer=OldFactorIndexToNew.begin();citer!=OldFactorIndexToNew.end();++citer) {
		for(std::vector<size_t>::const_iterator citer_i=(*factorTo)[citer->first].begin();citer_i!=(*factorTo)[citer->first].end();++citer_i) {
			if((*NodesOnMachine)[*citer_i]==k) {
				Nalpha[CumPosFactorsToNodes[citer->second]+tmpFactorAdd[citer->second]] = OldNodeIndexToNew.find(*citer_i)->second;
				NodePosInFactor[CumPosFactorsToNodes[citer->second]+tmpFactorAdd[citer->second]++] = citer_i - (*factorTo)[citer->first].begin();
			}
		}		
	}

	//generate c_i and hat_c_alpha
	std::vector<T> c_i(OldNodeIndexToNew.size());
	for(std::map<size_t, size_t>::const_iterator citer=OldNodeIndexToNew.begin();citer!=OldNodeIndexToNew.end();++citer) {
		c_i[citer->second] = (*in_c_i)[citer->first];
	}
	std::vector<T> hat_c_alpha(OldFactorIndexToNew.size());
	for(std::map<size_t, size_t>::const_iterator citer=OldFactorIndexToNew.begin();citer!=OldFactorIndexToNew.end();++citer) {
		hat_c_alpha[citer->second] = (*in_c_alpha)[citer->first]/Npalpha[citer->first].size();
	}

	//stack the node potentials
	std::vector<size_t> s_nodePotentialIX(OldNodeIndexToNew.size(), size_t(-1));
	std::vector<PotentialStruct<T>* > s_nodePotentials;
	std::vector<size_t> s_NodePotPosition(1,0);
	std::map<size_t,size_t> OldNodePotIndexToNew;

	std::vector<size_t>::iterator iter_s_nodePotIX = s_nodePotentialIX.begin();
	for(std::map<size_t,size_t>::const_iterator citer=OldNodeIndexToNew.begin();citer!=OldNodeIndexToNew.end();++citer, ++iter_s_nodePotIX) {
		size_t OldNodePotIX = (*nodePotentialIX)[citer->first];
		if(OldNodePotIX==size_t(-1)) {
			continue;
		}
		std::map<size_t,size_t>::iterator potIndex = OldNodePotIndexToNew.find(OldNodePotIX);
		if(potIndex==OldNodePotIndexToNew.end()) {
			OldNodePotIndexToNew.insert(std::make_pair<size_t,size_t>(OldNodePotIX, s_nodePotentials.size()));
			*iter_s_nodePotIX = s_nodePotentials.size();
			s_nodePotentials.push_back(&((*nodePotentials)[OldNodePotIX]));
			s_NodePotPosition.push_back((*nodePotentials)[OldNodePotIX].num_states[0] + s_NodePotPosition.back());
		} else {
			*iter_s_nodePotIX = potIndex->second;
		}
	}

	T* s_NodePotentials = new T[s_NodePotPosition.back()];
	std::vector<size_t>::iterator potPosIter = s_NodePotPosition.begin();
	for(typename std::vector<PotentialStruct<T>* >::iterator potIter=s_nodePotentials.begin();potIter!=s_nodePotentials.end();++potIter, ++potPosIter) {
		T* pot = s_NodePotentials + *potPosIter;
		memcpy(pot, (*potIter)->pot, (*(potPosIter+1)-*potPosIter)*sizeof(T));
	}

	//stack the factor potentials
	std::vector<size_t> s_factorPotentialIX(OldFactorIndexToNew.size(), size_t(-1));
	std::vector<std::pair<PotentialStruct<T>* , size_t> > s_factorPotentials;
	std::vector<size_t> s_FactorPotPosition(1,0);
	std::vector<size_t> s_FactorVariableSizePosition(1,0);

	std::vector<size_t>::iterator iter_s_factorPotIX = s_factorPotentialIX.begin();
	for(std::map<size_t,size_t>::const_iterator citer=OldFactorIndexToNew.begin();citer!=OldFactorIndexToNew.end();++citer, ++iter_s_factorPotIX) {
		size_t OldFactorPotIX = (*factorPotentialIX)[citer->first];

		std::pair<PotentialStruct<T>*, size_t> NewEntry = std::make_pair<PotentialStruct<T>*, size_t>(&((*factorPotentials)[OldFactorPotIX]), Npalpha[citer->first].size());

		bool found = false;
		for(typename std::vector<std::pair<PotentialStruct<T>* , size_t> >::iterator iter=s_factorPotentials.begin();iter!=s_factorPotentials.end();++iter) {
			if(*iter==NewEntry) {
				*iter_s_factorPotIX = iter - s_factorPotentials.begin();
				found = true;
				break;
			}
		}
		if(!found) {
			*iter_s_factorPotIX = s_factorPotentials.size();
			s_factorPotentials.push_back(NewEntry);

			size_t PotSz = 1;
			for(size_t m=0;m<(*factorPotentials)[OldFactorPotIX].num_variables;++m) {
				PotSz *= (*factorPotentials)[OldFactorPotIX].num_states[m];
			}

			s_FactorPotPosition.push_back(PotSz + s_FactorPotPosition.back());
			s_FactorVariableSizePosition.push_back((*factorPotentials)[OldFactorPotIX].num_variables + s_FactorVariableSizePosition.back());
		}
	}

	T* s_FactorPotentials = new T[s_FactorPotPosition.back()];
	size_t* s_FactorVariableSize = new size_t[s_FactorVariableSizePosition.back()];
	potPosIter = s_FactorPotPosition.begin();
	std::vector<size_t>::iterator potVarIter = s_FactorVariableSizePosition.begin();
	for(typename std::vector<std::pair<PotentialStruct<T>*,size_t> >::iterator potIter=s_factorPotentials.begin();potIter!=s_factorPotentials.end();++potIter, ++potPosIter, ++potVarIter) {
		T* pot = s_FactorPotentials + *potPosIter;
		for(size_t m=0;m<*(potPosIter+1)-*potPosIter;++m) {
			pot[m] = potIter->first->pot[m]/potIter->second;		//divide potential
		}
		size_t* var = s_FactorVariableSize + *potVarIter;
		memcpy(var, potIter->first->num_states, (*(potVarIter+1)-*potVarIter)*sizeof(size_t));
	}

	//which messages client should provide to server, i.e. which factors are divided onto multiple machines
	std::vector<size_t> s_DividedFactors(OldFactorIndexToNew.size()+1,size_t(-1));
	size_t lastSize = 0;
	for(std::map<size_t, size_t>::const_iterator citer=OldFactorIndexToNew.begin();citer!=OldFactorIndexToNew.end();++citer) {
		if(Npalpha[citer->first].size()>1) {
			s_DividedFactors[citer->second] = lastSize;
			size_t potIX = s_factorPotentialIX[citer->second];
			lastSize += s_FactorPotPosition[potIX+1] - s_FactorPotPosition[potIX];
		}
	}
	s_DividedFactors.back() = lastSize;

	//transmit all the information to the client
	std::vector<size_t> problemInfo;
	problemInfo.push_back(s_nodePotentialIX.size());
	problemInfo.push_back(s_factorPotentialIX.size());
	problemInfo.push_back(cBPparams->ClientIterations);		//how often to be iterated on the client side
	problemInfo.push_back(s_NodePotPosition.size());
	problemInfo.push_back(s_FactorPotPosition.size());
	problemInfo.push_back(s_FactorVariableSizePosition.size());

	if(sizeof(double)>sizeof(size_t)) {
		float epsilonTmp = float(cBPparams->epsilon);
		float athreshold = float(cBPparams->agreementThreshold);
		problemInfo.push_back(*((size_t*)&epsilonTmp));
		problemInfo.push_back(*((size_t*)&athreshold));
	} else {
		problemInfo.push_back(*((size_t*)&cBPparams->epsilon));
		problemInfo.push_back(*((size_t*)&cBPparams->agreementThreshold));
	}

	MPI::COMM_WORLD.Send((char*)&problemInfo[0], int(sizeof(size_t)*problemInfo.size()), MPI::CHAR, int(k+1), 0);
	MPI::COMM_WORLD.Send((char*)&s_nodePotentialIX[0], int(sizeof(size_t)*s_nodePotentialIX.size()), MPI::CHAR, int(k+1), 1);
	MPI::COMM_WORLD.Send((char*)&s_factorPotentialIX[0], int(sizeof(size_t)*s_factorPotentialIX.size()), MPI::CHAR, int(k+1), 2);
	MPI::COMM_WORLD.Send((char*)&CumPosNodesToFactors[0], int(sizeof(size_t)*CumPosNodesToFactors.size()), MPI::CHAR, int(k+1), 3);
	MPI::COMM_WORLD.Send((char*)&CumPosFactorsToNodes[0], int(sizeof(size_t)*CumPosFactorsToNodes.size()), MPI::CHAR, int(k+1), 4);
	MPI::COMM_WORLD.Send((char*)&Ni[0], int(sizeof(size_t)*Ni.size()), MPI::CHAR, int(k+1), 5);
	MPI::COMM_WORLD.Send((char*)&Nalpha[0], int(sizeof(size_t)*Nalpha.size()), MPI::CHAR, int(k+1), 6);
	MPI::COMM_WORLD.Send((char*)&s_NodePotSize[0], int(sizeof(size_t)*s_NodePotSize.size()), MPI::CHAR, int(k+1), 7);
	MPI::COMM_WORLD.Send((char*)&NodePosInFactor[0], int(sizeof(size_t)*NodePosInFactor.size()), MPI::CHAR, int(k+1), 8);
	MPI::COMM_WORLD.Send((char*)&c_i[0], int(sizeof(T)*c_i.size()), MPI::CHAR, int(k+1), 9);
	MPI::COMM_WORLD.Send((char*)&hat_c_alpha[0], int(sizeof(T)*hat_c_alpha.size()), MPI::CHAR, int(k+1), 10);
	MPI::COMM_WORLD.Send((char*)&s_NodePotPosition[0], int(sizeof(size_t)*s_NodePotPosition.size()), MPI::CHAR, int(k+1), 11);
	MPI::COMM_WORLD.Send((char*)s_NodePotentials, int(sizeof(T)*s_NodePotPosition.back()), MPI::CHAR, int(k+1), 12);
	MPI::COMM_WORLD.Send((char*)&s_FactorPotPosition[0], int(sizeof(size_t)*s_FactorPotPosition.size()), MPI::CHAR, int(k+1), 13);
	MPI::COMM_WORLD.Send((char*)&s_FactorVariableSizePosition[0], int(sizeof(size_t)*s_FactorVariableSizePosition.size()), MPI::CHAR, int(k+1), 14);
	MPI::COMM_WORLD.Send((char*)s_FactorPotentials, int(sizeof(T)*s_FactorPotPosition.back()), MPI::CHAR, int(k+1), 15);
	MPI::COMM_WORLD.Send((char*)s_FactorVariableSize, int(sizeof(size_t)*s_FactorVariableSizePosition.back()), MPI::CHAR, int(k+1), 16);
	MPI::COMM_WORLD.Send((char*)&s_DividedFactors[0], int(sizeof(size_t)*s_DividedFactors.size()), MPI::CHAR, int(k+1), 17);

	AllS_DividedFactors[k] = s_DividedFactors;
	AllS_OldNodeIndexToNew[k] = OldNodeIndexToNew;
	AllS_OldFactorIndexToNew[k] = OldFactorIndexToNew;
	AllS_CumPosFactorsToNodes[k] = CumPosFactorsToNodes;
	AllS_Nalpha[k] = Nalpha;
	AllS_NodePotSize[k] = s_NodePotSize;
	AllS_NodePosInFactor[k] = NodePosInFactor;

	delete [] s_FactorPotentials;
	delete [] s_FactorVariableSize;
	delete [] s_NodePotentials;
	
	return 0;
}

template <class T>
int Server<T>::NewIteration() {
	dualValue = T(0.0);
	primalValue = T(0.0);
	num_agreements = T(0.0);
	return 0;
}

template <class T>
int Server<T>::ReceiveMsgFromClient(int machine, std::vector<char*>& IdxMsg, std::vector<int>& IdxSz, std::vector<char*>& MsgMsg) {
	MPI::Status stat;

	MPI::COMM_WORLD.Probe(machine+1, 0, stat);
	int count = stat.Get_count(MPI::CHAR);
	char* Idx = new char[count];
	MPI::COMM_WORLD.Recv(Idx, count, MPI::CHAR, machine+1, 0, stat);
	IdxMsg.push_back(Idx);

	IdxSz.push_back(count);

	MPI::COMM_WORLD.Probe(machine+1, 1, stat);
	count = stat.Get_count(MPI::CHAR);
	char* ptr = new char[count];
	MPI::COMM_WORLD.Recv(ptr, count, MPI::CHAR, machine+1, 1, stat);
	MsgMsg.push_back(ptr);

	dualValue += *((T*)ptr);
	ptr += sizeof(T);
	primalValue += *((T*)ptr);
	ptr += sizeof(T);
	num_agreements += (T)(*((T*)ptr));
	ptr += sizeof(T);

	return 0;
}

template <class T>
int Server<T>::ComputePrimalDualAgreement() {
	std::cout << "Dual: " << dualValue << "\tPrimal: " << primalValue << "\tAgreements: " << num_agreements << std::endl;

	return 0;
}

template <class T>
int Server<T>::ComputeAndSendUpdates(std::vector<char*>& IdxMsg, std::vector<int>& IdxSz, std::vector<char*>& MsgMsg) {
	
	std::vector<T*> MsgToBeSent(m_numMachines, NULL);
	for(size_t k=0;k<m_numMachines;++k) {
		if(AllS_DividedFactors[k].back()>0) {
			MsgToBeSent[k] = new T[AllS_DividedFactors[k].back()];
		}
	}

	for(size_t alpha=0;alpha<Npalpha.size();++alpha) {
		if(Npalpha[alpha].size()>1) {		//this clique is divided
			size_t facPotIX = (*factorPotentialIX)[alpha];
			size_t numVarsInFactor = (*factorPotentials)[facPotIX].num_variables;
			size_t* prodSize = new size_t[numVarsInFactor + 1];
			prodSize[0] = 1;
			for(size_t i=0;i<numVarsInFactor;++i) {
				prodSize[i+1] = prodSize[i]*(*factorPotentials)[facPotIX].num_states[i];
			}

			T* n_sTOalpha = new T[prodSize[numVarsInFactor]];
			memset((char*)n_sTOalpha, 0, sizeof(T)*prodSize[numVarsInFactor]);

			for(std::set<size_t>::iterator s_iter=Npalpha[alpha].begin();s_iter!=Npalpha[alpha].end();++s_iter) {
				T* ns_array = (T*)MsgMsg[*s_iter];

				size_t alpha_machineLocal = AllS_OldFactorIndexToNew[*s_iter].find(alpha)->second;

				for(size_t iIX=AllS_CumPosFactorsToNodes[*s_iter][alpha_machineLocal];iIX<AllS_CumPosFactorsToNodes[*s_iter][alpha_machineLocal+1];++iIX) {
					size_t i_machineLocal=AllS_Nalpha[*s_iter][iIX];

					size_t pos = 3;		//dual, primal and agreement
					size_t* idxmsg = (size_t*)IdxMsg[*s_iter];
					for(size_t findPos=0;findPos<IdxSz[*s_iter]/sizeof(size_t);findPos+=2) {
						if(idxmsg[findPos]==alpha_machineLocal && idxmsg[findPos+1]==i_machineLocal) {
							break;
						}
						pos += AllS_NodePotSize[*s_iter][idxmsg[findPos+1]];
					}

					size_t nodepos = AllS_NodePosInFactor[*s_iter][iIX];
					size_t varsize = AllS_NodePotSize[*s_iter][i_machineLocal];

					for(size_t x_alpha=0;x_alpha<prodSize[numVarsInFactor];++x_alpha) {
						size_t var = (x_alpha/prodSize[nodepos])%varsize;

						n_sTOalpha[x_alpha] += *(ns_array + pos + var);
					}

				}
			}

			for(size_t x_alpha=0;x_alpha<prodSize[numVarsInFactor];++x_alpha) {
				n_sTOalpha[x_alpha] /= Npalpha[alpha].size();
			}

			//message for machine s
			for(std::set<size_t>::iterator s_iter=Npalpha[alpha].begin();s_iter!=Npalpha[alpha].end();++s_iter) {
				T* ns_array = (T*)MsgMsg[*s_iter];

				T* s_msg = new T[prodSize[numVarsInFactor]];
				memcpy((char*)s_msg, (char*)n_sTOalpha, sizeof(T)*prodSize[numVarsInFactor]);

				size_t alpha_machineLocal = AllS_OldFactorIndexToNew[*s_iter].find(alpha)->second;

				for(size_t iIX=AllS_CumPosFactorsToNodes[*s_iter][alpha_machineLocal];iIX<AllS_CumPosFactorsToNodes[*s_iter][alpha_machineLocal+1];++iIX) {
					size_t i_machineLocal=AllS_Nalpha[*s_iter][iIX];

					size_t pos = 3;		//dual, primal and agreement
					size_t* idxmsg = (size_t*)IdxMsg[*s_iter];
					for(size_t findPos=0;findPos<IdxSz[*s_iter]/sizeof(size_t);findPos+=2) {
						if(idxmsg[findPos]==alpha_machineLocal && idxmsg[findPos+1]==i_machineLocal) {
							break;
						}
						pos += AllS_NodePotSize[*s_iter][idxmsg[findPos+1]];
					}

					size_t nodepos = AllS_NodePosInFactor[*s_iter][iIX];
					size_t varsize = AllS_NodePotSize[*s_iter][i_machineLocal];

					for(size_t x_alpha=0;x_alpha<prodSize[numVarsInFactor];++x_alpha) {
						size_t var = (x_alpha/prodSize[nodepos])%varsize;

						s_msg[x_alpha] -= *(ns_array + pos + var);
					}
				}

				//now we have the appropriate message in s_msg
				memcpy((char*)(MsgToBeSent[*s_iter] + AllS_DividedFactors[*s_iter][alpha_machineLocal]), s_msg, sizeof(T)*prodSize[numVarsInFactor]);

				delete [] s_msg;
			}

			delete [] prodSize;
			delete [] n_sTOalpha;
		}
	}

	char note[] = "Update";
	for(int k=0;k<int(m_numMachines);++k) {
		MPI::COMM_WORLD.Send(note, 7, MPI::CHAR, k+1, 0);
		if(AllS_DividedFactors[k].back()>0) {
			MPI::COMM_WORLD.Send((char*)MsgToBeSent[k], int(sizeof(T)*AllS_DividedFactors[k].back()), MPI::CHAR, k+1, 1);
		} else {
			char nothing = '\0';
			MPI::COMM_WORLD.Send(&nothing, 1, MPI::CHAR, k+1, 1);
		}
	}

	for(size_t k=0;k<m_numMachines;++k) {
		if(AllS_DividedFactors[k].back()>0) {
			delete [] MsgToBeSent[k];
		}
	}

	return 0;
}

template <class T>
int Server<T>::SendBeliefRequest() {
	char note[] = "Belief";
	for(int k=0;k<int(m_numMachines);++k) {
		MPI::COMM_WORLD.Send(note, 7, MPI::CHAR, k+1, 0);
	}
	return 0;
}

template <class T>
int Server<T>::ReceiveBeliefs(const char *fn) {
	//dump beliefs
	MPI::Status stat;
	std::vector<char*> locBel(m_numMachines, NULL);
	for(int k=0;k<int(m_numMachines);++k) {
		MPI::COMM_WORLD.Probe(k+1, 0, stat);
		int count = stat.Get_count(MPI::CHAR);
		locBel[k] = new char[count];
		MPI::COMM_WORLD.Recv(locBel[k], count, MPI::CHAR, k+1, 0, stat);
	}

	std::ofstream ofs(fn, std::ios_base::out | std::ios_base::binary);

	for(size_t k=0;k<m_numNodes;++k) {
		size_t machineID = (*NodesOnMachine)[k];
		size_t newNodeIndex = AllS_OldNodeIndexToNew[machineID][k];

		size_t newPos = 0;
		for(size_t m=0;m<newNodeIndex;++m) {
			newPos += AllS_NodePotSize[machineID][m]*sizeof(T);
		}

		ofs.write(locBel[machineID] + newPos, sizeof(T)*AllS_NodePotSize[machineID][newNodeIndex]);
	}
	ofs.close();

	for(size_t k=0;k<m_numMachines;++k) {
		delete [] locBel[k];
	}
	locBel.clear();

	return 0;
}

template class Server<double>;
template class Server<float>;