/*  This file is part of distributed convex BP - http://www.alexander-schwing.de/
 *
 *  distributed convex BP is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  distributed convex BP is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with distributed convex BP.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Copyright (C) 2010-2011  Alexander Schwing  [aschwing _at_ inf _dot_ ethz _dot_ ch]
 */

#include "GeneralDefinitions.h"

#include <mpi.h>

#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include <cstdlib>
#include <string>
#include <cmath>
#include <algorithm>
#include <string.h>
#include <assert.h>

#include "Server.h"
#include "Client.h"

template <class T>
int ReadCBP(const char* fileName, std::vector<std::vector<size_t> > &nodeTo, std::vector<std::vector<size_t> > &factorTo, std::vector<size_t> &factorPotentialIX, std::vector<PotentialStruct<T> > &factorPotentials, std::vector<PotentialStruct<T> > &nodePotentials, std::vector<size_t> &nodePotentialIX, std::vector<size_t> &variableSize, std::vector<size_t> &NodesOnMachine, std::vector<T>& c_i, std::vector<T>& c_alpha) {
	std::cout << "Reading Potentials... ";

	std::ifstream ifs(fileName, std::ios_base::in | std::ios_base::binary);
	if(!ifs.is_open()) {
		std::cout << "File not open." << std::endl;
		return -1;
	}

	uint64 accuracy;
	ifs.read((char*)&accuracy, sizeof(uint64));
	assert(accuracy==sizeof(T));

	uint64 num_nodes, num_factors;

	ifs.read((char*)&num_nodes, sizeof(uint64));
	variableSize.assign(size_t(num_nodes), size_t(-1));
	for(uint64 k=0;k<num_nodes;++k) {
		uint64 tmp;
		ifs.read((char*)&tmp, sizeof(uint64));
		variableSize[size_t(k)] = size_t(tmp);
	}
	NodesOnMachine.assign(size_t(num_nodes), size_t(-1));
	for(uint64 k=0;k<num_nodes;++k) {
		uint64 tmp;
		ifs.read((char*)&tmp, sizeof(uint64));
		NodesOnMachine[size_t(k)] = size_t(tmp);
	}
	c_i.assign(size_t(num_nodes), T(1.0));
	ifs.read((char*)&c_i[0], sizeof(T)*size_t(num_nodes));

	nodePotentialIX.assign(size_t(num_nodes), size_t(-1));
	nodeTo.assign(size_t(num_nodes), std::vector<size_t>());

	ifs.read((char*)&num_factors, sizeof(uint64));
	std::map<uint64, size_t > FilePotIXTOInternalNodePotIX;
	std::map<uint64, size_t > FilePotIXTOInternalFactorPotIX;

	for(uint64 k=0;k<num_factors;++k) {
		uint64 num_vars, potIX;
		ifs.read((char*)&num_vars, sizeof(uint64));
		ifs.read((char*)&potIX, sizeof(uint64));

		std::vector<uint64> var_ix(size_t(num_vars), uint64(-1));
		if(num_vars==1) {
			ifs.read((char*)&var_ix[0], sizeof(uint64));

			std::map<uint64, size_t >::iterator curIX = FilePotIXTOInternalNodePotIX.find(potIX);
			if(curIX!=FilePotIXTOInternalNodePotIX.end()) {
				nodePotentialIX[size_t(var_ix[0])] = curIX->second;
			} else {
				size_t curIXval = FilePotIXTOInternalNodePotIX.size();
				nodePotentialIX[size_t(var_ix[0])] = curIXval;

				PotentialStruct<T> buf;
				buf.num_variables = 1;
				buf.num_states = new size_t[1];
				buf.num_states[0] = variableSize[size_t(var_ix[0])];
				buf.pot = new T[variableSize[size_t(var_ix[0])]];
				nodePotentials.push_back(buf);

				FilePotIXTOInternalNodePotIX.insert(std::make_pair<uint64, size_t>(potIX, curIXval));
			}
		} else {
			ifs.read((char*)&var_ix[0], sizeof(uint64)*size_t(num_vars));

			for(uint64 m=0;m<num_vars;++m) {
				nodeTo[size_t(var_ix[size_t(m)])].push_back(factorTo.size());
			}

			std::vector<size_t> tmp_var_ix(size_t(num_vars), size_t(-1));
			for(uint64 k=0;k<num_vars;++k) {
				tmp_var_ix[size_t(k)] = size_t(var_ix[size_t(k)]);
			}
			factorTo.push_back(tmp_var_ix);
			
			std::map<uint64, size_t>::iterator curIX = FilePotIXTOInternalFactorPotIX.find(potIX);
			if(curIX!=FilePotIXTOInternalFactorPotIX.end()) {
				factorPotentialIX.push_back(curIX->second);
			} else {
				size_t curIXval = FilePotIXTOInternalFactorPotIX.size();
				factorPotentialIX.push_back(curIXval);

				size_t potSize = 1;
				size_t *potSizes = new size_t[size_t(num_vars)];
				for(uint64 m=0;m<num_vars;++m) {
					potSizes[size_t(m)] = variableSize[size_t(var_ix[size_t(m)])];
					potSize *= potSizes[size_t(m)];					
				}
				PotentialStruct<T> buf;
				buf.num_variables = size_t(num_vars);
				buf.num_states = potSizes;
				buf.pot = new T[potSize];
				factorPotentials.push_back(buf);

				FilePotIXTOInternalFactorPotIX.insert(std::make_pair<uint64, size_t>(potIX, curIXval));
			}
		}
	}
	c_alpha.assign(factorTo.size(), T(1.0));
	ifs.read((char*)&c_alpha[0], sizeof(T)*factorTo.size());

	while(!ifs.eof()) {
		uint64 potIX;
		ifs.read((char*)&potIX, sizeof(uint64));

		if(ifs.eof()) {
			break;
		}

		std::map<uint64, size_t >::iterator pos = FilePotIXTOInternalNodePotIX.find(potIX);
		if(pos!=FilePotIXTOInternalNodePotIX.end()) {
			ifs.read((char*)nodePotentials[pos->second].pot, sizeof(T)*nodePotentials[pos->second].num_states[0]);
		} else {
			std::map<uint64, size_t >::iterator pos = FilePotIXTOInternalFactorPotIX.find(potIX);
			
			if(pos!=FilePotIXTOInternalFactorPotIX.end()) {
				size_t factorSize = 1;
				for(size_t k=0;k<factorPotentials[pos->second].num_variables;++k) {
					factorSize *= factorPotentials[pos->second].num_states[k];
				}
				
				ifs.read((char*)factorPotentials[pos->second].pot, sizeof(T)*factorSize);
			}
		}
	}

	ifs.close();

	std::cout << "done" << std::endl;
	return 0;
}

template <class T>
int ReadUAI(const char* fileName, std::vector<std::vector<size_t> > &nodeTo, std::vector<std::vector<size_t> > &factorTo, std::vector<size_t> &factorPotentialIX, std::vector<PotentialStruct<T> > &factorPotentials, std::vector<PotentialStruct<T> > &nodePotentials, std::vector<size_t> &nodePotentialIX, std::vector<size_t> &variableSize, std::vector<size_t> &NodesOnMachine) {
	std::cout << "Reading Potentials... ";

	size_t num_nodes, num_cliques;
	std::string NetType;

	std::ifstream ifs(fileName, std::ios_base::in);
	if(!ifs.is_open()) {
		std::cout << "File not open." << std::endl;
		return -1;
	}

	ifs >> NetType;
	if(NetType.compare("MARKOV")!=0) {
		std::cout << "The Network type is not MARKOV." << std::endl;
		return -1;
	}
	ifs >> num_nodes;
	variableSize.assign(num_nodes, size_t(-1));
	for(size_t k=0;k<num_nodes;++k) {
		ifs >> variableSize[k];
	}
	ifs >> num_cliques;

	size_t num_vars;
	nodeTo.assign(num_nodes, std::vector<size_t>());
	nodePotentialIX.assign(num_nodes, size_t(-1));
	std::vector<T*> storeLocation;
	std::vector<size_t*> storeSize;
	std::vector<size_t> storeVars;
	for(size_t k=0;k<num_cliques;++k) {
		ifs >> num_vars;
		
		std::vector<size_t> var_ix(num_vars);
		if(num_vars==1) {
			ifs >> var_ix[0];
			nodePotentialIX[var_ix[0]] = nodePotentials.size();
			PotentialStruct<T> buf;
			buf.num_variables = 1;
			buf.num_states = new size_t[1];
			buf.num_states[0] = variableSize[var_ix[0]];
			buf.pot = new T[variableSize[var_ix[0]]];
			nodePotentials.push_back(buf);
			storeLocation.push_back(buf.pot);
			storeSize.push_back(buf.num_states);
			storeVars.push_back(buf.num_variables);
		} else if(num_vars>=2) {
			for(size_t m=0;m<num_vars;++m) {
				ifs >> var_ix[m];
			}
			factorTo.push_back(var_ix);
			factorPotentialIX.push_back(factorPotentials.size());
			size_t potSize = 1;
			size_t *potSizes = new size_t[num_vars];
			for(size_t m=0;m<num_vars;++m) {
				nodeTo[var_ix[m]].push_back(factorPotentials.size());
				potSize *= variableSize[var_ix[m]];
				potSizes[m] = variableSize[var_ix[m]];
			}
			PotentialStruct<T> buf;
			buf.num_variables = num_vars;
			buf.num_states = potSizes;
			buf.pot = new T[potSize];
			factorPotentials.push_back(buf);
			storeLocation.push_back(buf.pot);
			storeSize.push_back(buf.num_states);
			storeVars.push_back(buf.num_variables);
		}
	}

	size_t sz;
	T potEntry;
	for(size_t k=0;k<storeLocation.size();++k) {
		size_t* prodState = new size_t[storeVars[k]+1];
		size_t* prodStateRev = new size_t[storeVars[k]+1];
		prodState[storeVars[k]] = 1;
		prodStateRev[0] = 1;
		for(size_t n=storeVars[k]-1;n>=0 && n<storeVars[k];--n) {
			prodState[n] = prodState[n+1]*storeSize[k][n];
			int revVar = -int(n)+int(storeVars[k]);
			prodStateRev[revVar] = prodStateRev[revVar-1]*storeSize[k][revVar-1];
		}

		ifs >> sz;
		assert(prodState[0]==sz);
		for(size_t m=0;m<sz;++m) {
			ifs >> potEntry;

			size_t* varState = new size_t[storeVars[k]];
			for(size_t n=storeVars[k]-1;n>=0 && n<storeVars[k];--n) {
				varState[n] = (m/prodState[n+1])%storeSize[k][n];
			}

			size_t newPos = 0;
			for(size_t n=0;n<storeVars[k];++n) {
				newPos += varState[n]*prodStateRev[n];
			}
			delete [] varState;

			storeLocation[k][newPos] = potEntry;
		}

		delete [] prodState;
		delete [] prodStateRev;
	}
	

	ifs.close();

	NodesOnMachine.assign(num_nodes, 0);
	//NodesOnMachine[2] = 1;
	//NodesOnMachine[3] = 1;

	std::cout << "done" << std::endl;

	return 0;
}

void ClearDataStructure(std::vector<char*>& MsgMsg,	std::vector<char*>& IdxMsg,	std::vector<int>& IdxSz) {
	for(size_t k=0;k<IdxMsg.size();++k) {
		delete [] IdxMsg[k];
		delete [] MsgMsg[k];
	}
	IdxMsg.clear();
	MsgMsg.clear();
	IdxSz.clear();
}

template <class T>
int Compute(struct cBP_Settings<double>& params, int numprocs, bool CBPFormat) {
	std::vector<std::vector<size_t> > nodeTo;					//the factors, node v is connected to: nodeTo[v].begin() ... nodeTo[v].end()
	std::vector<std::vector<size_t> > factorTo;					//the nodes, factor alpha is connected to: factor[alpha].begin() ... factor[alpha].end()
	std::vector<PotentialStruct<T> > nodePotentials;			//different node potentials
	std::vector<PotentialStruct<T> > factorPotentials;			//different factor potentials
	std::vector<size_t> nodePotentialIX;						//the node Potential of node v is stored in nodePotentials[nodePotentialIX[v]]
	std::vector<size_t> factorPotentialIX;						//the factor Potential of factor alpha is stored in factorPotentials[factorPotentialIX[alpha]]
	std::vector<size_t> variableSize;							//the number of states for random variable v is stored in variableSize[v]
	std::vector<size_t> NodesOnMachine;							//which nodes are assigned to which machine, i.e. partitioning nodes to machines
	std::vector<T> c_i;											//c_i of node v is c_i[v]
	std::vector<T> c_alpha;										//c_alpha of factor alpha is c_alpha[alpha]

	int res;
	if(params.fn==NULL) {
		res = ReadUAI("example.uai", nodeTo, factorTo, factorPotentialIX, factorPotentials, nodePotentials, nodePotentialIX, variableSize, NodesOnMachine);
		c_i.assign(nodeTo.size(), 1.0);
		c_alpha.assign(factorTo.size(), 1.0);
	} else {
		if(CBPFormat) {
			res = ReadCBP(params.fn, nodeTo, factorTo, factorPotentialIX, factorPotentials, nodePotentials, nodePotentialIX, variableSize, NodesOnMachine, c_i, c_alpha);
		} else {
			res = ReadUAI(params.fn, nodeTo, factorTo, factorPotentialIX, factorPotentials, nodePotentials, nodePotentialIX, variableSize, NodesOnMachine);
			c_i.assign(nodeTo.size(), 1.0);
			c_alpha.assign(factorTo.size(), 1.0);
		}
	}

	if(res!=0) {
		std::cout << "An error occured while reading the problem." << std::endl;
		MPI::COMM_WORLD.Abort(1);
		return 0;
	}

	if(int(*std::max_element(NodesOnMachine.begin(), NodesOnMachine.end()))+1!=numprocs-1) {
		std::cout << "Number of Machines does not fit specified problem." << std::endl;
		MPI::COMM_WORLD.Abort(1);
		return 0;
	}

	Server<T> SV(&nodeTo, &factorTo, &nodePotentials, &factorPotentials, &nodePotentialIX, &factorPotentialIX, &variableSize, &params, &NodesOnMachine, &c_i, &c_alpha);

	for(int k=1;k<numprocs;++k) {
		SV.SendProblemToClient(k-1);
	}

	std::vector<char*> MsgMsg;
	std::vector<char*> IdxMsg;
	std::vector<int> IdxSz;
	size_t MaxIteration = params.ServerIterations;

	CPrecisionTimer CTmr;
	CTmr.Start();
	for(size_t iteration=0;iteration<MaxIteration;++iteration) {
		SV.NewIteration();
		for(int k=1;k<numprocs;++k) {
			SV.ReceiveMsgFromClient(k-1, IdxMsg, IdxSz, MsgMsg);
		}
		SV.ComputePrimalDualAgreement();
		if(iteration<MaxIteration-1) {
			SV.ComputeAndSendUpdates(IdxMsg, IdxSz, MsgMsg);
			ClearDataStructure(MsgMsg, IdxMsg, IdxSz);
		} else {
			ClearDataStructure(MsgMsg, IdxMsg, IdxSz);
			break;
		}
	}
	std::cout << "Time for " << params.ServerIterations*params.ClientIterations << " iterations: " << CTmr.Stop() << "s" << std::endl;
	SV.SendBeliefRequest();
	if(params.OutputFile==NULL) {
		SV.ReceiveBeliefs("LocalBeliefs.txt");
	} else {
		SV.ReceiveBeliefs(params.OutputFile);
	}

	return 0;
}

void PrintHelp() {
	cBP_Settings<double> tmp;
	std::cout << "Invoke using e.g.:" << std::endl << std::endl;
	std::cout << "    mpiexec -n <NUM_PROCESSES> dcBP <OPTIONS>" << std::endl << std::endl;
	std::cout << "<NUMPROCESSES>:" << std::endl;
	std::cout << "This parameter needs to fit the partitions encoded in the graphical problem." << std::endl;
	std::cout << std::endl;
	std::cout << "<OPTIONS>:" << std::endl;
	std::cout << "  -a <val.a>\tThe agreement threshold, i.e. if \\forall i\\in N(\\alpha), x_i: abs(b_i(x_i) - \\sum_{x_alpha\\setminus x_i} b(x_\\alpha)) <= <val.a>, node belief and factor belief are treated to agree. This is important for computation of the primal value. The default for <val.a> is " << tmp.agreementThreshold << "." << std::endl;
	std::cout << std::endl;
	std::cout << "  -e <val.e>\tThe parameter epsilon. Its default value is \\epsilon = " << tmp.epsilon << "." << std::endl;
	std::cout << std::endl;
	std::cout << "  -c <val.c>\tThe number of iterations to be performed on the client side without exchanging information between the sub-problems. The default value is " << tmp.ClientIterations << "." << std::endl;
	std::cout << std::endl;
	std::cout << "  -s <val.s>\tThe number of iterations to be performed on the server side. The default value is " << tmp.ServerIterations << ". Hence, the total number of iterations is val.c*val.s." << std::endl;
	std::cout << std::endl;
	std::cout << "  -f <val.f>\tThe filename which defines the graphical model. The standard is 'example.uai'." << std::endl;
	std::cout << std::endl;
	std::cout << "  -o <val.o>\tThe filename used to store the local beliefs b_i(x_i). The standard is 'LocalBeliefs.txt'." << std::endl;
	std::cout << std::endl;
	std::cout << "  -h        \tDisplays this help." << std::endl;
	std::cout << std::endl << std::endl;
}

template <class T>
int ParseInput(int argc, char** argv, cBP_Settings<T>& params) {
	for(int k=1;k<argc;++k) {
		if(strcmp(argv[k], "-a")==0 && k+1!=argc) {
			params.agreementThreshold = T(atof(argv[++k]));
		} else if(strcmp(argv[k], "-e")==0 && k+1!=argc) {
			params.epsilon = T(atof(argv[++k]));
		} else if(strcmp(argv[k], "-c")==0 && k+1!=argc) {
			params.ClientIterations = size_t(atoi(argv[++k]));
		} else if(strcmp(argv[k], "-s")==0 && k+1!=argc) {
			params.ServerIterations = size_t(atoi(argv[++k]));
		} else if(strcmp(argv[k], "-f")==0 && k+1!=argc) {
			params.fn = argv[++k];
		} else if(strcmp(argv[k], "-o")==0 && k+1!=argc) {
			params.OutputFile = argv[++k];
		} else if(strcmp(argv[k], "-h")==0) {
			PrintHelp();
			return 1;
		}
	}
	return 0;
}

template <class T>
int ClientPart(int myid) {
	Client<T> CL;

	CL.ReceiveProblem(myid);		

	while(1) {
		CL.DoComputation();
		CL.SendMsg();
		int res = CL.ReceiveUpdateOrBeliefRequest();
		if(res==1) {
			break;
		}
	}
	return 0;
}

int main(int argc, char** argv) {
	
	MPI::Status stat;

	MPI::Init(argc, argv);
	int numprocs = MPI::COMM_WORLD.Get_size();
	int myid = MPI::COMM_WORLD.Get_rank();
	//int myid = 0, numprocs = 3;

	if(numprocs==1) {
		PrintHelp();
	} else {
		if(myid==0) {
			struct cBP_Settings<double> params;
			if(ParseInput(argc, argv, params)==1) {
				MPI::COMM_WORLD.Abort(1);
				return 0;
			}

			std::cout << "Parameters:" << std::endl;
			std::cout << "  Threshold:\t" << params.agreementThreshold << std::endl;
			std::cout << "  Epsilon:  \t" << params.epsilon << std::endl;
			std::cout << "  Client It:\t" << params.ClientIterations << std::endl;
			std::cout << "  Server It:\t" << params.ServerIterations << std::endl;
			if(params.fn==NULL) {
				std::cout << "  Filename: \t" << "example.uai" << std::endl;
			} else {
				std::cout << "  Filename: \t" << params.fn << std::endl;
			}
			if(params.OutputFile==NULL) {
				std::cout << "  Output: \t" << "LocalBeliefs.txt" << std::endl;
			} else {
				std::cout << "  Output: \t" << params.OutputFile << std::endl;
			}
			std::cout << std::endl;

			bool CBPFormat = false;
			if(!(params.fn==NULL) && strcmp(params.fn+strlen(params.fn)-4,".cbp")==0) {
				CBPFormat = true;
			}

			uint64 accuracy = 8;//double precision
			if(CBPFormat==true) {
				std::ifstream ifs(params.fn, std::ios_base::in | std::ios_base::binary);
				if(!ifs.is_open()) {
					std::cout << "File not open." << std::endl;
					MPI::COMM_WORLD.Abort(1);
					return 0;
				}
				ifs.read((char*)&accuracy, sizeof(uint64));
				ifs.close();
			}

			for(int k=1;k<numprocs;++k) {
				MPI::COMM_WORLD.Send((char*)&accuracy, int(sizeof(uint64)), MPI::CHAR, k,0);
			}

			if(accuracy==8) {
				Compute<double>(params, numprocs, CBPFormat);
			} else {
				Compute<float>(params, numprocs, CBPFormat);
			}

		} else {
			MPI::Status stat;
			uint64 accuracy;
			MPI::COMM_WORLD.Recv((char*)&accuracy, int(sizeof(uint64)), MPI::CHAR, 0, 0, stat);

			if(accuracy==8) {
				ClientPart<double>(myid);
			} else {
				ClientPart<float>(myid);
			}
		}
	}
	
	MPI::Finalize();
	return 0;
}